;; Gary Jones, October 1998.
;; -------------------------
;; Convert the mother or child utterances (sentences) into
;; phonemes. This file contains functions to access the filtered
;; speech files (from child or mother). The filter functions
;; are detailed in filter.lisp.  The functions here deal
;; with reading in the recoded files and creating an output
;; file of the sentences coded in phonemes (i.e. accessing the
;; lexicon). Each word in a sentence is accessed in
;; the lexicon which has the phonemic representation of the
;; word. Any word which is not in the lexicon is output
;; to a separate file - the phonemic representation
;; of these will need to be added to the lexicon.
;; Any sentences involving words which are not in the lexicon
;; are output to a separate file.
;; Note: This file uses the *hash-table* defined in lexicon.lisp, and
;; uses some of the functions defined in lexicon.lisp.
;;
;; CHANGE RECORD
;; Ini  Date      Change summary & functions added, deleted, or altered.
;; ---  --------  ------------------------------------------------------
;; GJ   29/10/98  Initial program.
;;                ------------------------------------------------------
;; GJ   02/11/98  Ensure access to the convert function uses strings
;;                that are in upper case. Only the parse-sentence
;;                function required changing.
;;                ------------------------------------------------------
;; GJ   09/11/98  Added convert-two-noun-word, and changed parse-sentence
;;                to call it. The function is used to get the phonemic
;;                representation of two-noun words like santa+claus.
;;                The phoneme representation of each of the nouns
;;                needs to be retrieved and then the representations
;;                need to be joined together as one representation.
;;                ------------------------------------------------------
;; GJ   18/11/98  Changed referneces for *default-directory* to
;;                *project-directory* as there will be more than one project.
;;                ------------------------------------------------------
;; GJ   18/11/98  Phoneme output file now has parentheses around each line
;;                so that the model can read it in as one big list.
;;                File processing (for learning etc) will then be quicker.
;;                Changed: change-sentences-to-phonemes and parse-sentence.
;;                ------------------------------------------------------
;; GJ   18/11/98  Added a line in the variables section to set the
;;                *project-directory*. Also took out constants because
;;                they were already in the general-variables file.
;;                ------------------------------------------------------
;; GJ   08/02/99  Added functions add-frequencies and print-frequencies
;;                which count up the word frequencies for words that get
;;                converted into phonemes and output to the phonemic
;;                output file.
;;                ------------------------------------------------------
;; GJ   27/02/99  Added a check of lisp-implementation-version to determine
;;                whether to load fasl files or pfsl (vsn 4+) files.
;;                ------------------------------------------------------
;; GJ   03/03/99  All lexicon related files will be loaded from
;;                lexicon-loader.lisp now, so the initial setting of
;;                *hard-drive*, and loads for logical-dir and 
;;                general-functions, can be commented out.
;;                ------------------------------------------------------
;; GJ   10/03/99  Added functions list-words-not-in-lexicon and
;;                create-extra-lexicon-words-file to automate the
;;                creation of the extra-lexicon-words file.
;;                ------------------------------------------------------
;; GJ   11/03/99  Before writing to output files, delete the file if it
;;                already exists. This is because overwrite in Lisp
;;                leaves the tail end of the previous file if the
;;                overwritten information is shorter.
;;                ------------------------------------------------------
;; ======== END OF CHANGE RECORD ========

;(defvar *hard-drive* "HD")
;(cond
;  ((member "4" (mapcar #'string (coerce (lisp-implementation-version) 'list))
;           :test #'equal)
;    (load (concatenate 'string  *hard-drive*
;                       ":logical-dir-compatibility.pfsl"))
;    (load (concatenate 'string  *project-directory*
;                                "general-functions.pfsl")))
;  (t
;    (load (concatenate 'string  *hard-drive*
;                       ":logical-dir-compatibility.fasl"))
;    (load (concatenate 'string  *project-directory*
;                                "general-functions.fasl"))))
;(def-logical-directory "home;" *project-directory*)


;;---------------------- VARIABLES -----------------------------;;

(defvar *non-lexicon-words* nil "Output file of all the words in sentences
                                 that do not exist in the lexicon.")
(defvar *non-lexicon-sentences* nil "Output file of all the sentences which
                                     contain words that do not exist in
                                     the lexicon.")
(defvar *words-not-known* nil "All of the words not known by the lexicon.")
(defvar *extra-lexicon-word-file* "extra-lexicon-words"
        "The file containing all of the extra lexicon words that arise from
         the word not being in the lexicon because of the difference in
         coding schemes between our input and other input.")
(defvar *extra-lexicon-words* '()
        "All of the extra lexicon words that arise from the word not
         being in the lexicon because of the difference in coding schemes
         between our input and other input.")
(defvar *word-list* '()
        "The word frequencies for the utterance words for which
         there is a phonemic representation for all the words
         in the sentence i.e. don't count up words in a sentence
         which gets dropped.")
(defvar *word-frequencies* '()
        "The list of words for which there is a phonemic representation
         for all the words in the sentence i.e. don't list words in
         a sentence which gets dropped.")



;;---------------------- FUNCTIONS -----------------------------;;

(defun change-sentences-to-phonemes ()
  "Change the sentences in a file to a list of the phonemes
   representing the words in the sentence."
  (let ((temp nil))
    (setf *directory* *project-directory*)
    (setf *words-not-known* '())
    (setf *word-frequencies* '())
    (setf *word-list* '())
    (get-all-non-lexicon-words)
    (format t "~%Enter input directory (default: ~A): " *directory*)
    (setf temp (read-line))
    (if (> (length temp) 0)
      (progn
        (setf *directory* temp)
        (setf temp nil)))
    (format t "~%Enter input file (enter=quit) : ")
    (setf temp (read-line))
    (loop while (> (length temp) 0) do
      (setf *input-file*
              (concatenate 'string *directory*
                                   temp))
      (setf *output-file*
              (concatenate 'string *input-file*
                                   *phoneme-file-suffix*))
      (delete-if-exists *output-file*)
      (setf *non-lexicon-words* 
              (concatenate 'string *input-file*
                                   *non-lexicon-word-file-suffix*))
      (delete-if-exists *non-lexicon-words*)
      (setf *non-lexicon-sentences*
              (concatenate 'string *input-file*
                                   *non-lexicon-sentence-file-suffix*))
      (delete-if-exists *non-lexicon-sentences*)
      (with-open-file (input *input-file*
                             :direction :input)
        (with-open-file (output *output-file*
                                :direction :output)
          (with-open-file (wds-output *non-lexicon-words*
                                      :direction :output)
            (with-open-file (snt-output *non-lexicon-sentences*
                                        :direction :output)
              (princ *open-parentheses* output)
              (do ((a-line (read-line input nil 'eof nil)
                           (read-line input nil 'eof nil)))
                ((eq a-line 'eof) 'nil)
                  (if (string-contains-alphabetic-p a-line)
                    (parse-sentence a-line output wds-output snt-output)))
              (princ *close-parentheses* output)))))

      (delete-if-exists (concatenate 'string *input-file*
                                             ".freq"))
      (with-open-file (freq-output (concatenate 'string *input-file*
                                                        ".freq")
                                   :direction :output)
        (print-frequencies freq-output))
      
      (format t "~%Enter input file (enter=quit) : ")
      (setf temp (read-line)))))


(defun mass-change ()
  "Do a mass change of sentences to phonemes."
  (setf *directory* 
        "HD:Syntax project:syntax:maternal-files:Filtered mother files:")
  (get-all-non-lexicon-words)
  (dolist (temp '("amy.mot.filt" "bethan.mot.filt"
                  "david.mot.filt" "georgie.mot.filt"
                  "james.mot.filt" "natalie.mot.filt"
                  "adam.mot.filt" "conor.mot.filt"
                  "jack.mot.filt" "lucy.mot.filt"
                  "rois.mot.filt" "will.mot.filt"))
    (setf *words-not-known* '())
    (setf *word-frequencies* '())
    (setf *word-list* '())
    (setf *input-file*
            (concatenate 'string *directory*
                                 temp))
    (setf *output-file*
            (concatenate 'string *input-file*
                                 *phoneme-file-suffix*))
    (delete-if-exists *output-file*)
    (setf *non-lexicon-words* 
            (concatenate 'string *input-file*
                                 *non-lexicon-word-file-suffix*))
    (delete-if-exists *non-lexicon-words*)
    (setf *non-lexicon-sentences*
            (concatenate 'string *input-file*
                                 *non-lexicon-sentence-file-suffix*))
    (delete-if-exists *non-lexicon-sentences*)
    (with-open-file (input *input-file*
                           :direction :input)
      (with-open-file (output *output-file*
                              :direction :output
                              :if-exists :overwrite)
        (with-open-file (wds-output *non-lexicon-words*
                                    :direction :output
                                    :if-exists :overwrite)
          (with-open-file (snt-output *non-lexicon-sentences*
                                      :direction :output
                                      :if-exists :overwrite)
            (princ *open-parentheses* output)
            (do ((a-line (read-line input nil 'eof nil)
                         (read-line input nil 'eof nil)))
              ((eq a-line 'eof) 'nil)
                (if (string-contains-alphabetic-p a-line)
                  (parse-sentence a-line output wds-output snt-output)))
            (princ *close-parentheses* output)))))
    
    (delete-if-exists (concatenate 'string *input-file*
                                             ".freq"))
    (with-open-file (freq-output (concatenate 'string *input-file*
                                                      ".freq")
                                 :direction :output)
      (print-frequencies freq-output))))


(defun parse-sentence (a-line output-file words-file sentence-file)
  "Parse the sentence. If each word can be converted into
   phonemes then convert it and place the conversion in the
   output file. If a word cannot be converted, then put it
   in the word file and the sentence is put in the sentence file."
  (let ((phoneme-rep nil) (word nil) (converted-sentence nil)
        (sentence-ok t) (sentence nil) (stored-sentence nil))
    ;; The hash table has all entries in upper case, so the
    ;; access to the hash table has to be in upper case.
    (setf sentence (coerce-string-into-list (string-upcase a-line)))

    ;; Store the sentence for word frequency use.
    (setf stored-sentence sentence)

    ;; NB Loop for every word in the sentence so that we get
    ;; a full list of the words that aren't in the lexicon.
    (loop while sentence do
      (setf word (string (first sentence)))
      (setf sentence (rest sentence))
      (setf word (convert-if-extraneous-word word))
      (cond
        ((search "+" word)
          (setf phoneme-rep (convert-multi-noun-word word)))
        (t
          (setf phoneme-rep (convert word))))
;(princ word) (princ phoneme-rep) (terpri)
      (cond
        (phoneme-rep
          (setf converted-sentence (concatenate 'string
                                                converted-sentence
                                                (list-to-string phoneme-rep)
                                                " "
                                                *word-separator*
                                                " ")))
        (t
          (setf sentence-ok nil)
          (cond
            ((not (member word *words-not-known* :test #'equal))
              (setf *words-not-known* (append *words-not-known*
                                              (list word)))
              (princ word words-file)
              (terpri words-file)
              (setf sentence-ok nil))))))
    (if sentence-ok
      (progn
        (add-frequencies stored-sentence)
        (princ *open-parentheses* output-file)
        (princ converted-sentence output-file)
        (princ *close-parentheses* output-file)
        (terpri output-file))
      (progn
        (princ a-line sentence-file)
        (terpri sentence-file)))))


(defun convert-multi-noun-word (word)
  "Some words will be two nouns e.g. santa+claus. This wants 
   to remain as one word, but the phonemic 
   representation of each word needs to be found before doing
   the combination."
  (let ((position nil) (noun nil) (phoneme-rep t) (word-rep nil))
    (setf word (string-upcase word))
    (setf position (search "+" word))
;(princ position)
    (loop while (and phoneme-rep position) do
      (setf noun (subseq word 0 position))
      (setf word (subseq word
                         (+ position 1)
                         (length word)))
      (setf phoneme-rep (convert noun))
      (cond
        (phoneme-rep
          (setf word-rep (append word-rep phoneme-rep))
          (setf position (search "+" word)))
        (t 
          (setf word-rep nil))))
    ;; Now do the final word.
    (cond
      ((and phoneme-rep      ;; i.e. all previous words were ok
            (convert word))  ;; i.e. the last word is ok
        (setf phoneme-rep (convert word))
        (setf word-rep (append word-rep phoneme-rep)))
      (t
        (setf word-rep nil)))
    word-rep))


(defun get-all-non-lexicon-words ()
  "Read all the non-lexicon words and their phonemic representations
   into a variable to be searched."
  (let ((word-line nil))
    (setf *extra-lexicon-words* '())
    (with-open-file (input (concatenate 'string
                                        *project-directory*
                                        *extra-lexicon-word-file*)
                           :direction :input)
      (do ((a-line (read-line input nil 'eof nil)
                   (read-line input nil 'eof nil)))
          ((eq a-line 'eof) 'nil)
        (cond
           ((string-contains-alphabetic-p a-line)
            (setf word-line (coerce-string-into-list a-line))
            (setf *extra-lexicon-words* (append
                                          *extra-lexicon-words*
                                          (list (list (first word-line)
                                                      (second word-line))))))
          (t nil))))))


(defun convert-if-extraneous-word (word)
  "Determine whether a word is part of the list of words
   that are not in the lexicon but should have a phonemic
   representation. If so, give the proper word otherwise
   output the word given."
  (let ((temp-list *extra-lexicon-words*) (word-found nil))
    (loop while (and temp-list (not word-found)) do
;(princ temp-list)
      (cond
        ((equal (first (first temp-list)) word)
          (setf word-found t)
          (setf word (second (first temp-list))))
        (t
          (setf temp-list (rest temp-list)))))
    word))


(defun add-frequencies (sentence)
  "Add the frequencies for the words in the sentence."
  (let ((word nil) (position 0))
    (loop while sentence do
      (setf word (car sentence))
      (cond
        ((not (equal (string word) *word-separator*))
          (cond
            ((member word *word-list* :test #'equal)
              (setf position (- (length *word-list*)
                                (length (member word *word-list* :test #'equal))))
              (setf (nth position *word-frequencies*)
                      (+ (nth position *word-frequencies*) 1)))
            (t
              (setf *word-list* (append *word-list*
                                        (list word)))
              (setf *word-frequencies* (append *word-frequencies*
                                               (list 1)))))))
      (setf sentence (rest sentence)))))


(defun print-frequencies (outfile)
  "Print all the word frequencies."
  (let ((tot 0.0))
    (dolist (n *word-frequencies*)
      (setf tot (+ tot n)))
    (format outfile "Entries: ~A; Average frequency: ~A~%"
                    (length *word-frequencies*)
                    (round (/ tot (length *word-frequencies*)))))
  (dotimes (x (length *word-list*))
    (format outfile "~A ~A~%" (nth x *word-list*) (nth x *word-frequencies*))))


(defun list-words-not-in-lexicon ()
  "Create a file of all the words in the input that
   are not in the lexicon, based on the non-lexicon word
   files created for each mothers input."
  (setf *directory* (concatenate 'string "HD:Syntax project:syntax:"
                                 "maternal-files:Phonemic mother files:"))
  (let  ((words '()))
    (dolist (temp '("amy.mot.filt" "bethan.mot.filt"
                    "david.mot.filt" "georgie.mot.filt"
                    "james.mot.filt" "natalie.mot.filt"
                    "adam.mot.filt" "conor.mot.filt"
                    "jack.mot.filt" "lucy.mot.filt"
                    "rois.mot.filt" "will.mot.filt"))
      (with-open-file (input (concatenate 'string *directory* temp ".non-lex-wd")
                             :direction :input)
        (do ((a-line (read-line input nil 'eof nil)
                     (read-line input nil 'eof nil)))
            ((eq a-line 'eof) 'nil)
          (cond
            ((not (member (first (coerce-string-into-list a-line))
                          words :test #'string-equal))
              (setf words (append words 
                                  (list
                                    (first (coerce-string-into-list a-line))))))))))
    (delete-if-exists (concatenate 'string *directory*
                                           "Full non-lex words"))
    (with-open-file (output (concatenate 'string *directory*
                                                 "Full non-lex words")
                                 :direction :output)
      (dolist (word words)
        (format output "~A~%" word)))))


(defun create-extra-lexicon-words-file ()
  "Go through the file of words that are not in the lexicon
   and ask the user for an alternative representation or
   spelling of the words to place in the extra-lexicon-words
   file."
  (let ((non-lex-wd nil) (non-lex-wds '()) (alt-word nil))
    (with-open-file (input 
  "HD:syntax project:Syntax:Maternal-files:Phonemic mother files:Full non-lex words"
                             :direction :input)
        (do ((a-line (read-line input nil 'eof nil)
                     (read-line input nil 'eof nil)))
            ((eq a-line 'eof) 'nil)
          (setf non-lex-wd (coerce-string-into-list a-line))
          (cond
            ((= (length non-lex-wd) 1)
              (setf non-lex-wds (append non-lex-wds 
                                        (list (first non-lex-wd))))))))
    (get-all-non-lexicon-words)
    (dolist (word non-lex-wds)
      (cond
        ((string-equal word (convert-if-extraneous-word word))
          (format t "~%Alternative for ~A (enter=none)? " word)
          (setf alt-word (read-line))
          (cond
            ((and (> (length alt-word) 0)
                  (string-contains-alphabetic-p alt-word))
              (setf *extra-lexicon-words*
                      (append *extra-lexicon-words*
                              (list (list word (string-upcase alt-word))))))))))
    (delete-if-exists (concatenate 'string
                                   *project-directory*
                                   *extra-lexicon-word-file*))
    (with-open-file (output (concatenate 'string
                                         *project-directory*
                                         *extra-lexicon-word-file*)
                                 :direction :output)
      (dolist (extra-words *extra-lexicon-words*)
        (format output "~A ~A~%" (first extra-words) (second extra-words))))))