;; G. Jones, October 1998.
;; -----------------------
;; A group of general functions that will be used by various different
;; files. Placing the functions in one file means that any
;; changes to the functions will mean the functionality changes
;; everywhere the function is used.
;;
;; CHANGE RECORD
;; Ini  Date      Change summary & functions added, deleted, or altered.
;; ---  --------  ------------------------------------------------------
;; GJ   30/10/98  Initial program.
;;                ------------------------------------------------------
;; GJ   30/10/98  The coerce-string-into-list function does not actually
;;                work if the string contains only one item; now it is
;;                changed to return a one item list too.
;;                ------------------------------------------------------
;; GJ   30/10/98  Added list-to-string function.
;;                ------------------------------------------------------
;; GJ   02/11/98  Forced to change coerce-string-into-list as it
;;                a. fell over for single item strings
;;                b. fell over when presented with strings containing tabs.
;;                ------------------------------------------------------
;; GJ   03/12/98  Changed coerce-string-into-list so that tab separated
;;                words appear as two different items in the list. Also
;;                return characters are ignored, because they led to 
;;                filter.lisp falling over (the repeat codes could appear
;;                at the beginning of a line if the previous word
;;                contained an accidental return character).
;;                ------------------------------------------------------
;; GJ   04/12/98  Removing the return characters from the line (change
;;                of 03/12/98) is not necessary, but is left in anyway.
;;                The falling over was because the input contained an
;;                accidental return character in the middle of a line,
;;                just before a repeat code. I've left in the removal of
;;                returns anyway.
;;                ------------------------------------------------------
;; GJ   09/12/98  Added a sort function written as part of the ToN
;;                simulation for sorting when you have lists within lists.
;;                ------------------------------------------------------
;; GJ   10/12/98  Added calculate-number-value-from-string, which calculates
;;                the numeric value of a string, always returning a
;;                floating point number.
;;                ------------------------------------------------------
;; GJ   20/12/98  Added probability-true which returns true if a random
;;                number is within a specific probability. This is for
;;                node creation when nodes are created probabilistically.
;;                ------------------------------------------------------
;; GJ   15/01/99  Added function flatten from ToN simulation file
;;                code-files.lisp.
;; ---  --------  ------------------------------------------------------
;; GJ   20/01/99  Added a check of lisp-implementation-version to determine
;;                whether to load fasl files or pfsl (vsn 4+) files.
;; ---  --------  ------------------------------------------------------
;; GJ   26/01/99  Added function remove-numerics-in-string which removes
;;                all numerics from the passed string. This is primarily
;;                for removing stress from phonemes.
;; ---  --------  ------------------------------------------------------
;; GJ   08/02/99  Changed references of HD to *hard-drive*.
;; ---  --------  ------------------------------------------------------
;; GJ   12/02/99  Added functions string-is-all-alphabetics and
;;                return-string-between-columns and 
;;                coerce-chars-in-string-into-list.
;; ---  --------  ------------------------------------------------------
;; GJ   02/03/99  Added list-to-string-no-spaces.
;; ---  --------  ------------------------------------------------------
;; GJ   11/03/99  Added functions write-file-header and delete-if-exists.
;; ======== END OF CHANGE RECORD ========


(load (concatenate 'string *hard-drive*
                   ":syntax project:syntax:code:loop.lisp"))

(if (member "4" (mapcar #'string (coerce (lisp-implementation-version) 'list))
    :test #'equal)
  (load (concatenate 'string *hard-drive*
                     ":syntax project:syntax:code:general-variables.pfsl"))
  (load (concatenate 'string *hard-drive*
                     ":syntax project:syntax:code:general-variables.fasl")))


;;---------------------- FUNCTIONS -----------------------------;;

;(defun coerce-string-into-list (a-string)
;  (let ((list-string (ccl::READ-LIST-FROM-STRING a-string)))
;    (if (listp list-string)
;      list-string
;      (list list-string))))

(defun save-app (name)
  (gc)
  (save-application (concatenate 'string "home;" name)
                    :creator :gary
                    :init-file "home;lexicon"))


(defun write-file-header (file-stream filepath header)
  "Write a file header with author, date and the file
   path, and also a header describing the file."
  (multiple-value-bind (tmp mins hours day month year) (get-decoded-time)
    (format file-stream ";; File created on ~A/~A/~A at ~A:~A.~%"
                        day month year hours mins)
    (format file-stream ";; Filepath: ~A~%" filepath)
    (format file-stream ";; ~A~%" header)
    (format file-stream "~%"))
)


(defun delete-if-exists (fname)
  "Delete a given file if it already exists."
  (cond
    ((probe-file fname)
      (delete-file fname))))


(defun coerce-string-into-list (a-string)
  "Make a list out of a string. Individual items are
   assumed to be separated by whitespace or tabs."
  (let ((index 0) (list-element "") (final-list '())
        (string-length (length a-string)))
    (loop while (not (>= index string-length)) do
      (loop while (and (not (>= index string-length))
                       (or (equal (char a-string index) #\Space)
                           (equal (char a-string index) #\Tab)
                           (equal (char a-string index) #\Return))) do
        (setf index (+ index 1)))
      (loop while (and (not (>= index string-length))
                       (not (equal (char a-string index) #\Space))
                       (not (equal (char a-string index) #\Tab))
                       (not (equal (char a-string index) #\Return))) do
        (setf list-element (concatenate 'string list-element
                                                (string (char a-string index))))
        (setf index (+ index 1)))
      (if (> (length list-element) 0)
        (progn
          (setf final-list (append final-list (list list-element)))
          (setf list-element ""))))
    final-list))


(defun coerce-chars-in-string-into-list (a-string)
  "Make a list out of a string, treating each non-tab and
   non-return character as a separate element in the list."
  (let ((index 0) (final-list '())
        (string-length (length a-string)))
    (loop while (not (>= index string-length)) do
      (loop while (and (not (>= index string-length))
                       (or (equal (char a-string index) #\Tab)
                           (equal (char a-string index) #\Return))) do
        (setf index (+ index 1)))
      (cond
        ((< index string-length)
          (setf final-list (append final-list (list (string (char a-string index)))))
          (setf index (+ index 1)))))
    final-list))


(defun coerce-string-into-list-of-two-lists (a-string)
  (let ((a-list (coerce-string-into-list a-string)))
    (list (first a-list) (rest a-list))))


(defun alphabetic-character (ch)
  "Return true if the character is alphabetic."
  (or (and (>= (char-int ch) *ascii-char-little-a*)
           (<= (char-int ch) *ascii-char-little-z*))
      (and (>= (char-int ch) *ascii-char-big-a*)
           (<= (char-int ch) *ascii-char-big-z*))))


(defun convert-word-to-letter-list (a-word)
  "Convert the word to a list of the letters."
  (mapcar #'string (coerce a-word 'list)))


(defun string-contains-parentheses-p (a-string)
  "Return true if the string contains any parentheses."
  (let ((string-list (mapcar #'string (coerce a-string 'list))))
    (or (member *open-parentheses* string-list :test #'equal)
        (member *close-parentheses* string-list :test #'equal))))


(defun string-contains-alphabetic-p (a-string)
  "Return true if the string contains any letters."
  (let ((string-list (mapcar #'string (coerce a-string 'list))))
    (loop while (and string-list
                     (not (alphabetic-character (char (car string-list) 0)))) do
      (setf string-list (cdr string-list)))
    string-list))


(defun string-is-all-alphabetics-p (a-string)
  "Return true if the string contains only alphabetic characters."
  (let ((string-list (mapcar #'string (coerce a-string 'list))))
    (loop while (and string-list
                     (alphabetic-character (char (car string-list) 0))) do
      (setf string-list (cdr string-list)))
    (not string-list)))


(defun string-contains-full-stop-p (a-string)
  "Return true if the string contains any parentheses."
  (let ((string-list (mapcar #'string (coerce a-string 'list))))
    (member *full-stop* string-list :test #'equal)))


(defun string-contains-numerics-p (a-string)
  "Return true if the string contains any numerics."
  (let ((index 0)
        (line-length (length a-string)))
    (loop while (and (< index line-length)
                     (or (< (char-int (char a-string index)) *ascii-char-zero*)
                         (> (char-int (char a-string index)) *ascii-char-nine*))) do
      (setf index (+ index 1)))
    (if (>= index line-length)
      nil
      t)))


(defun string-contains-quote-p (a-string)
  "Return true if the string contains any quote characters."
  (let ((string-list (convert-word-to-letter-list a-string)))
    (member *quote* string-list :test #'equal)))


(defun change-quote-symbols (a-line)
  "Return true if the string contains any quote characters."
  (if (string-contains-quote-p a-line)
    (let ((index 0)
          (line-length (length a-line))
          (quote-length (length *quote*))
          (new-line ""))
      (loop while (< index line-length) do
        (if (string-equal a-line *quote*
                          :start1 index
                          :end1   (+ index quote-length))
          (setf new-line (format nil "~A~A" new-line *new-quote-symbol*))
          (setf new-line (format nil "~A~A" new-line (char a-line index))))
        (setf index (+ index 1)))
      new-line)
    a-line))


(defun remove-numerics-in-string (a-line)
  "Remove all numerics in a string."
  (if (string-contains-numerics-p a-line)
    (let ((index 0) (new-line "")
          (line-length (length a-line)))
      (loop while (< index line-length) do
        (if (or (< (char-int (char a-line index)) *ascii-char-zero*)
                (> (char-int (char a-line index)) *ascii-char-nine*))
          (setf new-line (format nil "~A~A" new-line (char a-line index))))
        (setf index (+ index 1)))
      new-line)
    a-line))


(defun get-first-word (a-line)
  "Get the first word in a line of text. The first word is taken
   to be the first string beginning with an alphabetic character,
   and ending at the first space character after the string."
  (let ((index 0)
        (first-word "")
        (line-length (length a-line)))
    (loop while (and (not (>= index line-length))
                     (not (alphabetic-character (char a-line index)))) do
      (setf index (+ index 1)))

    (loop while (and (not (>= index line-length))
                     (not (char-equal (char a-line index) #\Space))) do
      (setf first-word (format nil "~A~A" first-word (char a-line index)))
      (setf index (+ index 1)))
    first-word))


(defun list-to-string (list)
  "Turn a list of elements into a string of the 
   elements, separated by spaces."
  (let ((string ""))
    (dolist (item list)
      (if (equal string "")
        (setf string (format nil "~A" item))
        (setf string (format nil "~A ~A"
                       string
                       (format nil "~A" item)))))
    string)
)


(defun list-to-string-no-spaces (list)
  "Turn a list of elements into a string of the 
   elements."
  (let ((string ""))
    (dolist (item list)
      (if (equal string "")
        (setf string (format nil "~A" item))
        (setf string (format nil "~A~A"
                       string
                       (format nil "~A" item)))))
    string)
)


(defun return-string-between-columns (a-string col1 col2)
  "Return the string that lies between (and including)
   the columns given. The columns assume that the text
   starts on column 1."
  (cond
    ((and (> col1 0) (<= col2 (length a-string)))
      (subseq a-string (- col1 1) col2))))


(defun sort-list (a-list &key (order-function nil) (within-list-item nil))
  "Sort a list, using the order-function to decide how the sort
   should work, and the within-list-item function to decide which
   items to sort if there are lists within lists."
  (let ((temp-list (cdr a-list)) (first-item (car a-list)) (new-list '()))
    (loop while (cdr a-list) do
      (loop while temp-list do
        (if (funcall order-function (funcall within-list-item (car temp-list))
                                    (funcall within-list-item first-item))
          (setf first-item (car temp-list)))
        (setf temp-list (cdr temp-list)))
      (setf a-list (remove first-item a-list))
      (setf new-list (append new-list (list first-item)))
      (setf first-item (car a-list))
      (setf temp-list (cdr a-list)))
    ;; Now add the last item from the original list.
    (setf new-list (append new-list (list (first a-list))))
    new-list)
)


(defun calculate-number-value-from-string (number)
  "Turn a text number into its actual number. Note this also 
   applies to floating point numbers, and always returns a
   floating point number."
  (let ((index 0) (fp-num 0.0) (dot-pos nil) (non-numeric nil)
        (num-length (- (length number) 1)))

    ;; If nothing was input, set the non-numeric flag.
    (if (equal (length number) 0)
      (setf non-numeric t))

    ;; Find the decimal point, if there is one.
    (loop while (and (not (> index num-length))
                     (not (char-equal (char number index) #\.))) do
      (setf index (+ index 1)))

    (setf dot-pos index)
    (setf index 0)
    
    (loop while (and (not (>= index dot-pos))
                     (not non-numeric)) do
      (setf fp-num (+ fp-num
                      (* (- (char-int (char number index)) 48)
                         (expt 10 (- dot-pos index 1)))))
      (if (or (< (char-int (char number index)) 48)
              (> (char-int (char number index)) 57))
        (setf non-numeric t))
      (setf index (+ index 1)))

    ;; If we've got a dot, do the floating point part.
    (if (not (>= dot-pos num-length))
      (progn
        ;; Skip the dot.
        (setf index (+ index 1))
        (loop while (and (not (> index num-length))
                         (not non-numeric)) do
          (setf fp-num (+ fp-num
                          (* (- (char-int (char number index)) 48)
                             (expt 10 (- dot-pos index)))))
          (if (or (< (char-int (char number index)) 48)
                  (> (char-int (char number index)) 57))
            (setf non-numeric t))
          (setf index (+ index 1)))))

      (if non-numeric
        nil
        fp-num))
)


(defun probability-true (probability)
  "Return true if the randomly generated number is
   lower than the probability passed. For example, the
   random number could be 0.72 and the probability 0.8,
   which would return true."
  (<= (random 1.0) probability))


(defun flatten (list)
  "Flatten the lists within lists to be just a single list."
  (if (cdr list)
    (if (and (car list) (listp (car list)))
      (append (flatten (car list)) (flatten (cdr list)))
      (if (car list)
        (append (list (car list)) (flatten (cdr list)))
        (flatten (cdr list))))
    (if (and (car list) (listp (car list)))
      (flatten (car list))
      (if (car list)
        list
        '())))
)